@@ -23,8 +23,8 @@
 /* Type declarations		*/
 /*------------------------------*/
 
-void load_font(XFontStruct **);
-void createGC(Window, GC *, XFontStruct *);
+static void load_font(XFontStruct **);
+static void createGC(Window, GC *, XFontStruct *);
 
 /*----------------------------------*/
 /* Global variables for X11 drawing */
@@ -61,7 +61,7 @@ void highlight(int x, int y) {
     // If Obs2[] at x, y is a source or dest, don't highlight
     // Do this only for layer 0;  it doesn't have to be rigorous. 
     for (i = 0; i < Num_layers; i++) {
-	Pr = &Obs2[i][OGRID(x, y, i)];
+	Pr = &OBS2VAL(x, y, i);
 	if (Pr->flags & (PR_SOURCE | PR_TARGET)) return;
     }
 
@@ -100,7 +100,7 @@ void highlight_source() {
 	for (x = 0; x < NumChannelsX[i]; x++) {
 	    xspc = (x + 1) * spacing - hspc;
 	    for (y = 0; y < NumChannelsY[i]; y++) {
-		Pr = &Obs2[i][OGRID(x, y, i)];
+		Pr = &OBS2VAL(x, y, i);
 		if (Pr->flags & PR_SOURCE) {
 		    yspc = height - (y + 1) * spacing - hspc;
 		    XFillRectangle(dpy, win, gc, xspc, yspc,
@@ -136,7 +136,7 @@ void highlight_dest() {
 	for (x = 0; x < NumChannelsX[i]; x++) {
 	    xspc = (x + 1) * spacing - hspc;
 	    for (y = 0; y < NumChannelsY[i]; y++) {
-		Pr = &Obs2[i][OGRID(x, y, i)];
+		Pr = &OBS2VAL(x, y, i);
 		if (Pr->flags & PR_TARGET) {
 		    yspc = height - (y + 1) * spacing - hspc;
 		    XFillRectangle(dpy, win, gc, xspc, yspc,
@@ -175,11 +175,10 @@ void highlight_starts(POINT glist) {
 /* Highlight mask (in tan)		*/
 /*--------------------------------------*/
 
-void highlight_mask() {
+void highlight_mask(void) {
 
     int xspc, yspc, hspc;
     int x, y;
-    u_char value;
 
     if (RMask == NULL) return;
 
@@ -192,7 +191,7 @@ void highlight_mask() {
     for (x = 0; x < NumChannelsX[0]; x++) {
 	xspc = (x + 1) * spacing - hspc;
 	for (y = 0; y < NumChannelsY[0]; y++) {
-	    XSetForeground(dpy, gc, brownvector[RMask[OGRID(x, y, 0)]]);
+	    XSetForeground(dpy, gc, brownvector[RMASK(x, y)]);
 	    yspc = height - (y + 1) * spacing - hspc;
 	    XFillRectangle(dpy, win, gc, xspc, yspc, spacing, spacing);
 	}
@@ -204,13 +203,11 @@ void highlight_mask() {
 /* Draw a map of obstructions and pins		*/
 /*----------------------------------------------*/
 
-void
+static void
 map_obstruction()
 {
     int xspc, yspc, hspc;
-    int i, x, y, n, norm;
-    u_char *Congestion;
-    u_char value, maxval;
+    int i, x, y;
 
     hspc = spacing >> 1;
 
@@ -220,7 +217,7 @@ map_obstruction()
 	for (x = 0; x < NumChannelsX[i]; x++) {
 	    xspc = (x + 1) * spacing - hspc;
 	    for (y = 0; y < NumChannelsY[i]; y++) {
-		if (Obs[i][OGRID(x, y, i)] & NO_NET) {
+		if (OBSVAL(x, y, i) & NO_NET) {
 		    yspc = height - (y + 1) * spacing - hspc;
 		    XFillRectangle(dpy, buffer, gc, xspc, yspc,
 				spacing, spacing);
@@ -235,7 +232,7 @@ map_obstruction()
 	for (x = 0; x < NumChannelsX[i]; x++) {
 	    xspc = (x + 1) * spacing - hspc;
 	    for (y = 0; y < NumChannelsY[i]; y++) {
-		if (Nodesav[i][OGRID(x, y, i)] != NULL) {
+		if (NODESAV(x, y, i) != NULL) {
 		    yspc = height - (y + 1) * spacing - hspc;
 		    XFillRectangle(dpy, buffer, gc, xspc, yspc,
 				spacing, spacing);
@@ -249,7 +246,7 @@ map_obstruction()
 /* Draw a map of actual route congestion	*/
 /*----------------------------------------------*/
 
-void
+static void
 map_congestion()
 {
     int xspc, yspc, hspc;
@@ -267,12 +264,12 @@ map_congestion()
 	for (x = 0; x < NumChannelsX[i]; x++) {
 	    for (y = 0; y < NumChannelsY[i]; y++) {
 		value = (u_char)0;
-		n = Obs[i][OGRID(x, y, i)];
+		n = OBSVAL(x, y, i);
 		if (n & ROUTED_NET) value++;
 		if (n & BLOCKED_MASK) value++;
 		if (n & NO_NET) value++;
 		if (n & PINOBSTRUCTMASK) value++;
-		Congestion[OGRID(x, y, 0)] += value;
+		CONGEST(x, y) += value;
 	    }
 	}
     }
@@ -280,7 +277,7 @@ map_congestion()
     maxval = 0;
     for (x = 0; x < NumChannelsX[0]; x++) {
 	for (y = 0; y < NumChannelsY[0]; y++) {
-	    value = Congestion[OGRID(x, y, 0)];
+	    value = CONGEST(x, y);
 	    if (value > maxval) maxval = value;
 	}
     }
@@ -290,7 +287,7 @@ map_congestion()
     for (x = 0; x < NumChannelsX[0]; x++) {
 	xspc = (x + 1) * spacing - hspc;
 	for (y = 0; y < NumChannelsY[0]; y++) {
-	    XSetForeground(dpy, gc, bluevector[norm * Congestion[OGRID(x, y, 0)]]);
+	    XSetForeground(dpy, gc, bluevector[norm * CONGEST(x, y)]);
 	    yspc = height - (y + 1) * spacing - hspc;
 	    XFillRectangle(dpy, buffer, gc, xspc, yspc, spacing, spacing);
 	}
@@ -304,7 +301,7 @@ map_congestion()
 /* Draw a map of route congestion estimated from net bounding boxes	*/
 /*----------------------------------------------------------------------*/
 
-void
+static void
 map_estimate()
 {
     NET net;
@@ -334,13 +331,13 @@ map_estimate()
 
 	for (x = net->xmin; x < net->xmax; x++)
 	    for (y = net->ymin; y < net->ymax; y++)
-		Congestion[OGRID(x, y, 0)] += density;
+		CONGEST(x, y) += density;
     }
 
     maxval = 0.0;
     for (x = 0; x < NumChannelsX[0]; x++) {
 	for (y = 0; y < NumChannelsY[0]; y++) {
-	    density = Congestion[OGRID(x, y, 0)];
+	    density = CONGEST(x, y);
 	    if (density > maxval) maxval = density;
 	}
     }
@@ -350,7 +347,7 @@ map_estimate()
     for (x = 0; x < NumChannelsX[0]; x++) {
 	xspc = (x + 1) * spacing - hspc;
 	for (y = 0; y < NumChannelsY[0]; y++) {
-	    value = (int)(norm * Congestion[OGRID(x, y, 0)]);
+	    value = (int)(norm * CONGEST(x, y));
 	    XSetForeground(dpy, gc, bluevector[value]);
 	    yspc = height - (y + 1) * spacing - hspc;
 	    XFillRectangle(dpy, buffer, gc, xspc, yspc, spacing, spacing);
@@ -367,7 +364,6 @@ map_estimate()
 
 void draw_net(NET net, u_char single, int *lastlayer) {
 
-    int i, xspc, yspc;
     int layer;
     SEG seg;
     ROUTE rt;
@@ -422,16 +418,105 @@ void draw_net(NET net, u_char single, int *lastlayer) {
     }
 }
 
+/*--------------------------------------*/
+/* Draw one unrouted net on the display	*/
+/*--------------------------------------*/
+
+static void
+draw_net_nodes(NET net) {
+
+    NODE node;
+    SEG bboxlist = NULL; /* bbox list of all the nodes in the net */
+    SEG lastbbox, bboxit;
+    DPOINT tap;
+    int first, w, h, n;
+
+    /* Compute bbox for each node and draw it */
+    for (node = net->netnodes, n = 0; node != NULL; node = node->next, n++) {
+        if (bboxlist == NULL) {
+            lastbbox = bboxlist = (SEG)malloc(sizeof(struct seg_));
+        }
+        else {
+            lastbbox->next = (SEG)malloc(sizeof(struct seg_));
+            lastbbox = lastbbox->next;
+        }
+        lastbbox->next = NULL;
+        for (tap = node->taps, first = TRUE;
+             tap != NULL;
+             tap = tap->next, first = FALSE) {
+            if (first) {
+                lastbbox->x1 = lastbbox->x2 = tap->gridx;
+                lastbbox->y1 = lastbbox->y2 = tap->gridy;
+            }
+            else {
+                lastbbox->x1 = MIN(lastbbox->x1, tap->gridx);
+                lastbbox->x2 = MAX(lastbbox->x2, tap->gridx);
+                lastbbox->y1 = MIN(lastbbox->y1, tap->gridy);
+                lastbbox->y2 = MAX(lastbbox->y2, tap->gridy);
+            }
+        }
+
+        /* Convert to X coordinates */
+        lastbbox->x1 = spacing * (lastbbox->x1 + 1);
+        lastbbox->y1 = height - spacing * (lastbbox->y1 + 1);
+        lastbbox->x2 = spacing * (lastbbox->x2 + 1);
+        lastbbox->y2 = height - spacing * (lastbbox->y2 + 1);
+
+        /* Draw the bbox */
+        w = lastbbox->x2 - lastbbox->x1;
+        h = lastbbox->y1 - lastbbox->y2;
+        XDrawRectangle(dpy, buffer, gc,
+                       lastbbox->x1, lastbbox->y1, w, h
+        );
+    }
+
+    /* if net->numnodes == 1 don't draw a wire */
+    if (n == 2) {
+        XDrawLine(
+            dpy, buffer, gc,
+            (bboxlist->x1 + bboxlist->x2)/2, (bboxlist->y1 + bboxlist->y2)/2,
+            (lastbbox->x1 + lastbbox->x2)/2, (lastbbox->y1 + lastbbox->y2)/2
+        );
+    }
+    else if (n > 2) {
+        /* Compute center point */
+        POINT midpoint = (POINT)malloc(sizeof(struct point_));
+        midpoint->x1 = midpoint->y1 = 0;
+
+        for (bboxit = bboxlist; bboxit != NULL; bboxit = bboxit->next) {
+            midpoint->x1 += (bboxit->x1 + bboxit->x2)/2;
+            midpoint->y1 += (bboxit->y1 + bboxit->y2)/2;
+        }
+        midpoint->x1 /= n;
+        midpoint->y1 /= n;
+
+        for (bboxit = bboxlist; bboxit != NULL; bboxit = bboxit->next) {
+            XDrawLine(
+                dpy, buffer, gc,
+                midpoint->x1, midpoint->y1,
+                (bboxit->x1 + bboxit->x2)/2, (bboxit->y1 + bboxit->y2)/2
+            );
+        }
+
+        free(midpoint);
+    }
+
+    for (bboxit = bboxlist; bboxit != NULL; bboxit = lastbbox) {
+        lastbbox = bboxit->next;
+        free(bboxit);
+    }
+}
+
+
 /*--------------------------------------*/
 /* Graphical display of the layout	*/
 /*--------------------------------------*/
 
 void draw_layout() {
 
-    int lastlayer, xspc, yspc, hspc;
-    int i, x, y;
+    int lastlayer;
+    int i;
     NET net;
-    NETLIST nlist;
 
     if (dpy == NULL) return;
     else if (buffer == (Pixmap)NULL) return;
@@ -460,7 +545,7 @@ void draw_layout() {
     // Draw all nets, much like "emit_routes" does when writing
     // routes to the DEF file.
 
-    if ((mapType & DRAW_MASK) == DRAW_ROUTES) {
+    if ((mapType & DRAW_ROUTES) != 0) {
 	lastlayer = -1;
 	for (i = 0; i < Numnets; i++) {
 	    net = Nlnets[i];
@@ -468,6 +553,20 @@ void draw_layout() {
 	}
     }
 
+    // Draw unrouted nets
+
+    if ((mapType & DRAW_UNROUTED) != 0) {
+        XSetForeground(dpy, gc, blackpix);
+        for (i = 0; i < Numnets; i++) {
+            net = Nlnets[i];
+            if (strcmp(net->netname, gndnet) != 0
+                && strcmp(net->netname, vddnet) != 0
+                &&net->routes == NULL) {
+                draw_net_nodes(net);
+            }
+        }
+    }
+
     /* Copy double-buffer onto display window */
     XCopyArea(dpy, buffer, win, gc, 0, 0, width, height, 0, 0);
 }
@@ -488,7 +587,7 @@ int GUI_init(Tcl_Interp *interp)
    tktop = Tk_MainWindow(interp);
    if (tktop == NULL) {
       tcl_printf(stderr, "No Top-level Tk window available. . .\n");
-      return;
+      return TCL_ERROR;
    }
 
    qrouterdrawwin = (char *)Tcl_GetVar(interp, "drawwindow", TCL_GLOBAL_ONLY);
@@ -572,16 +671,8 @@ int GUI_init(Tcl_Interp *interp)
 
 /*----------------------------------------------------------------*/
 
-int QuitCallback(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *objv[])
-{
-   exit(0);
-   return TCL_OK;	// Statement not reached
-}
-
-/*----------------------------------------------------------------*/
-
-void load_font(XFontStruct **font_info)
+static void
+load_font(XFontStruct **font_info)
 {
    char *fontname = "9x15";
 
@@ -595,7 +686,8 @@ void load_font(XFontStruct **font_info)
 
 /*----------------------------------------------------------------*/
 
-void createGC(Window win, GC *gc, XFontStruct *font_info)
+static void
+createGC(Window win, GC *gc, XFontStruct *font_info)
 {
    unsigned long valuemask = 0; /* ignore XGCvalues and use defaults */
    XGCValues values;
@@ -668,7 +760,6 @@ int recalc_spacing()
 
 void resize(Tk_Window tkwind, int locwidth, int locheight)
 {
-   Window window;
 
    if ((locwidth == 0) || (locheight == 0)) return;
 
