@@ -337,7 +337,7 @@ u_char
 LefParseEndStatement(FILE *f, char *match)
 {
     char *token;
-    int keyword, words;
+    int keyword;
     char *match_name[2];
 
     match_name[0] = match;
@@ -613,7 +613,7 @@ LefFindLayerNum(char *token)
  */
 
 int
-LefGetMaxLayer()
+LefGetMaxLayer(void)
 {
     int maxlayer = -1;
     LefList lefl;
@@ -646,7 +646,6 @@ double
 LefGetRouteKeepout(int layer)
 {
     LefList lefl;
-    double dist;
 
     lefl = LefFindLayerByNum(layer);
     if (lefl) {
@@ -718,22 +717,48 @@ LefGetRouteOffset(int layer)
  * because the center can be on a half-grid position; so,
  * return half the value obtained.
  *
- * To-do:  Differentiate between X and Y vias when doing
- * checkerboard via patterning.
+ * This routine always uses a horizontally oriented via if
+ * available.  See the specific LefGetXYViaWidth() routine
+ * for differentiation between via orientations.
  *------------------------------------------------------------
  */
 
 double
 LefGetViaWidth(int base, int layer, int dir)
+{
+   return LefGetXYViaWidth(base, layer, dir, 0);
+}
+
+/*
+ *------------------------------------------------------------
+ * The base routing used by LefGetViaWidth(), with an
+ * additional argument that specifies which via orientation
+ * to use, if an alternative orientation is available.  This
+ * is necessary for doing checkerboard via patterning and
+ * for certain standard cells with ports that do not always
+ * fit one orientation of via.
+ *------------------------------------------------------------
+ */
+
+double
+LefGetXYViaWidth(int base, int layer, int dir, int orient)
 {
     DSEG lrect;
     LefList lefl;
     double width;
+    char **viatable;
 
-    lefl = LefFindLayer(ViaX[base]);
+    viatable = (orient == 1) ? ViaY : ViaX;
+
+    lefl = LefFindLayer(*(viatable + base));
+    if (!lefl) {
+	viatable = (orient == 1) ? ViaX : ViaY;
+	lefl = LefFindLayer(*(viatable + base));
+	viatable = (orient == 1) ? ViaY : ViaX;
+    }
     if (!lefl) {
 	if (base == Num_layers)
-	    lefl = LefFindLayer(ViaX[base - 1]);
+	    lefl = LefFindLayer(*(viatable + base - 1));
     }
     if (lefl) {
 	if (lefl->lefClass == CLASS_VIA) {
@@ -772,7 +797,10 @@ LefGetRouteSpacing(int layer)
     lefl = LefFindLayerByNum(layer);
     if (lefl) {
 	if (lefl->lefClass == CLASS_ROUTE) {
-	    return lefl->info.route.spacing->spacing;
+	    if (lefl->info.route.spacing)
+		return lefl->info.route.spacing->spacing;
+	    else
+		return 0.0;
 	}
     }
     return MIN(PitchX[layer], PitchY[layer]) / 2.0;
@@ -1742,8 +1770,8 @@ LefReadMacro(f, mname, oscale)
     /* Initial values */
     pinNum = 0;
     has_size = FALSE;
-    lefBBox.x1 = 0.0;
-    lefBBox.y1 = 0.0;
+    lefBBox.x2 = lefBBox.x1 = 0.0;
+    lefBBox.y2 = lefBBox.y1 = 0.0;
 
     while ((token = LefNextToken(f, TRUE)) != NULL)
     {
@@ -1956,11 +1984,11 @@ LefReadLayerSection(f, lname, mode, lefl)
     LefList lefl;		/* pointer to layer info  */
 {
     char *token, *tp;
-    int keyword, typekey, entries, i;
+    int keyword, typekey = -1, entries, i;
     struct seg_ viaArea;
     int curlayer = -1;
     double dvalue, oscale;
-    lefSpacingRule *newrule, *testrule;
+    lefSpacingRule *newrule = NULL, *testrule;
 
     /* These are defined in the order of CLASS_* in lefInt.h */
     static char *layer_type_keys[] = {
@@ -2239,15 +2267,17 @@ LefReadLayerSection(f, lname, mode, lefl)
  *------------------------------------------------------------
  */
 
-enum lef_sections {LEF_VERSION = 0, LEF_NAMESCASESENSITIVE,
-	LEF_PROPERTYDEFS, LEF_UNITS, LEF_SECTION_LAYER,
-	LEF_SECTION_VIA, LEF_SECTION_VIARULE,
+enum lef_sections {LEF_VERSION = 0,
+	LEF_BUSBITCHARS, LEF_DIVIDERCHAR, LEF_MANUFACTURINGGRID,
+	LEF_USEMINSPACING, LEF_CLEARANCEMEASURE,
+	LEF_NAMESCASESENSITIVE, LEF_PROPERTYDEFS, LEF_UNITS,
+	LEF_SECTION_LAYER, LEF_SECTION_VIA, LEF_SECTION_VIARULE,
 	LEF_SECTION_SPACING, LEF_SECTION_SITE, LEF_PROPERTY,
 	LEF_NOISETABLE, LEF_CORRECTIONTABLE, LEF_IRDROP,
 	LEF_ARRAY, LEF_SECTION_TIMING, LEF_EXTENSION, LEF_MACRO,
 	LEF_END};
 
-void
+int
 LefRead(inName)
     char *inName;
 {
@@ -2256,14 +2286,20 @@ LefRead(inName)
     char *token;
     char tsave[128];
     int keyword, layer;
+    int oprecis = 100;	// = 1 / manufacturing grid (microns)
     float oscale;
-    double xydiff;
+    double xydiff, ogrid;
     LefList lefl;
     DSEG grect;
     GATE gateginfo;
 
     static char *sections[] = {
 	"VERSION",
+	"BUSBITCHARS",
+	"DIVIDERCHAR",
+	"MANUFACTURINGGRID",
+	"USEMINSPACING",
+	"CLEARANCEMEASURE",
 	"NAMESCASESENSITIVE",
 	"PROPERTYDEFINITIONS",
 	"UNITS",
@@ -2295,7 +2331,7 @@ LefRead(inName)
     {
 	Fprintf(stderr, "Cannot open input file: ");
 	perror(filename);
-	return;
+	return 0;
     }
 
     if (Verbose > 0) {
@@ -2317,9 +2353,17 @@ LefRead(inName)
 	switch (keyword)
 	{
 	    case LEF_VERSION:
+	    case LEF_BUSBITCHARS:
+	    case LEF_DIVIDERCHAR:
+	    case LEF_CLEARANCEMEASURE:
+	    case LEF_USEMINSPACING:
+	    case LEF_NAMESCASESENSITIVE:
 		LefEndStatement(f);
 		break;
-	    case LEF_NAMESCASESENSITIVE:
+	    case LEF_MANUFACTURINGGRID:
+		token = LefNextToken(f, TRUE);
+		if (sscanf(token, "%lg", &ogrid) == 1)
+		    oprecis = (int)((1.0 / ogrid) + 0.5);
 		LefEndStatement(f);
 		break;
 	    case LEF_PROPERTYDEFS:
@@ -2523,4 +2567,5 @@ LefRead(inName)
 	    }
 	}
     }
+    return oprecis;
 }
