@@ -13,6 +13,7 @@
 
 #define _GNU_SOURCE	// for strcasestr(), see man page
 
+#include <ctype.h>
 #include <stdio.h>
 #include <math.h>
 #include <stdlib.h>
@@ -54,6 +55,7 @@ int     JogCost = 10;              // Cost of 1 grid off-direction jog
 int     XverCost = 4;              // Cost of a crossover
 int     BlockCost = 25;            // Cost of a crossover when node has
 				   // only one tap point
+int	OffsetCost = 50;	   // Cost per micron of a node offset
 int 	ConflictCost = 50;	   // Cost of shorting another route
 				   // during the rip-up and reroute stage
 
@@ -70,7 +72,7 @@ char    *ViaY[MAX_LAYERS];
 /*--------------------------------------------------------------*/
 
 void
-post_config()
+post_config(void)
 {
     int i, h, v;
 
@@ -99,15 +101,21 @@ post_config()
     else if (v == -1) v = h;
 
     for (i = 0; i < Num_layers; i++) {
-       if (PitchX[i] != 0.0 && PitchX[i] != PitchX[v])
+       if (PitchX[i] != 0.0 && PitchX[i] != PitchX[v]) {
 	  Fprintf(stderr, "Multiple vertical route layers at different"
-		" pitches.  I cannot handle this!  Using pitch %g\n",
-		PitchX[v]);
+		" pitches.  Using smaller pitch %g, will route on"
+		" 1-of-N tracks if necessary.\n",
+		PitchX[i]);
+	  PitchX[v] = PitchX[i];
+       }
        PitchX[i] = PitchX[v];
-       if (PitchY[i] != 0.0 && PitchY[i] != PitchY[h])
+       if (PitchY[i] != 0.0 && PitchY[i] != PitchY[h]) {
 	  Fprintf(stderr, "Multiple horizontal route layers at different"
-		" pitches.  I cannot handle this!  Using pitch %g\n",
-		PitchY[h]);
+		" pitches.  Using smaller pitch %g, will route on"
+		" 1-of-N tracks if necessary.\n",
+		PitchY[i]);
+	  PitchY[h] = PitchY[i];
+       }
        PitchY[i] = PitchY[h];
     }
 
@@ -132,14 +140,13 @@ int read_config(FILE *fconfig, int is_info)
     char carg;
     double darg, darg2, darg3, darg4;
     char sarg[MAX_LINE_LEN];
-    size_t len = 0;
     char   line[MAX_LINE_LEN];
     char  *lineptr;
     STRING dnr;               // Do Not Route nets
     STRING cn;                // critical nets
     STRING strl;
-    GATE   gateinfo;          // gate information, pin location, etc
-    DSEG   grect, drect;
+    GATE   gateinfo = NULL;   // gate information, pin location, etc
+    DSEG   drect;
 
     if (Firstcall) {
 	for (i = 0; i < MAX_LAYERS; i++) {
@@ -291,6 +298,9 @@ int read_config(FILE *fconfig, int is_info)
 	    OK = 1; XverCost = iarg;
 	}
 
+	if ((i = sscanf(lineptr, "route offset cost %d", &iarg)) == 1) {
+	    OK = 1; OffsetCost = iarg;
+	}
 	if ((i = sscanf(lineptr, "route block cost %d", &iarg)) == 1) {
 	    OK = 1; BlockCost = iarg;
 	}
