@@ -78,6 +78,7 @@ typedef unsigned long  u_long;
 
 typedef struct scalerec_ {
    int    iscale;
+   int	  mscale;
    double oscale;
 } ScaleRec;
 
@@ -115,6 +116,7 @@ struct proute_ {        // partial route
 #define PR_PRED_D	0x06		// Predecessor is down
 
 #define PR_PROCESSED	0x08		// Tag to avoid visiting more than once
+#define PR_NO_EVAL	0x08		// Used only for making calls to eval_pt()
 #define PR_CONFLICT	0x10		// Two nets collide here during stage 2
 #define PR_SOURCE	0x20		// This is a source node
 #define PR_TARGET	0x40		// This is a target node
@@ -192,6 +194,16 @@ struct route_ {
   u_char flags;         // See below for flags
 };
 
+/* Structure used to hold nodes, saved nodes, and stub/offset info */
+
+typedef struct nodeinfo_ *NODEINFO;
+
+struct nodeinfo_ {
+   NODE  nodesav;
+   NODE  nodeloc;
+   float stub;
+};
+
 /* Definitions for flags in struct route_ */
 
 #define RT_OUTPUT	0x1	// Route has been output
@@ -337,7 +349,8 @@ struct routeinfo_ {
 
 #define DRAW_NONE	0x0	// Draw only the background map
 #define DRAW_ROUTES	0x4	// Draw routes on top of background map
-#define DRAW_MASK	0x4
+#define DRAW_UNROUTED   0x8     // Draw unrouted nets on top of background map
+#define DRAW_MASK	0xc
 
 // Mask types (values other than 255 are interpreted as "slack" value)
 #define MASK_MINIMUM	(u_char)0	// No slack
@@ -376,11 +389,20 @@ extern u_char *RMask;
 extern u_int  *Obs[MAX_LAYERS];		// obstructions by layer, y, x
 extern PROUTE *Obs2[MAX_LAYERS]; 	// working copy of Obs 
 extern float  *Obsinfo[MAX_LAYERS];	// temporary detailed obstruction info
-extern float  *Stub[MAX_LAYERS];	// stub route distances to pins
-extern NODE   *Nodeloc[MAX_LAYERS];	// nodes are attached to grid points
-					// for reverse lookup
-extern NODE   *Nodesav[MAX_LAYERS];	// copy of Nodeloc used for restoring
-					// Nodeloc after net rip-up
+extern NODEINFO Nodeinfo[MAX_LAYERS];	// stub route distances to pins and
+					// pointers to node structures.
+
+#define NODEIPTR(x, y, l) (&(Nodeinfo[l][OGRID(x, y, l)]))
+#define NODESAV(x, y, l) (Nodeinfo[l][OGRID(x, y, l)].nodesav)
+#define NODELOC(x, y, l) (Nodeinfo[l][OGRID(x, y, l)].nodeloc)
+#define STUBVAL(x, y, l) (Nodeinfo[l][OGRID(x, y, l)].stub)
+#define OBSINFO(x, y, l) (Obsinfo[l][OGRID(x, y, l)])
+#define OBSVAL(x, y, l)  (Obs[l][OGRID(x, y, l)])
+#define OBS2VAL(x, y, l) (Obs2[l][OGRID(x, y, l)])
+
+#define RMASK(x, y)      (RMask[OGRID(x, y, 0)])
+#define CONGEST(x, y)	 (Congestion[OGRID(x, y, 0)])
+
 extern DSEG  UserObs;			// user-defined obstruction layers
 
 extern u_char needblock[MAX_LAYERS];
@@ -412,59 +434,34 @@ extern char *gndnet;
 
 /* Function prototypes */
 
-extern int    set_num_channels();
-extern int    allocate_obs_array();
+int    set_num_channels(void);
+int    allocate_obs_array(void);
+int    countlist(NETLIST net);
+int    runqrouter(int argc, char *argv[]);
 
-void   check_variable_pitch(int, int *, int *);
-
-NET    getnettoroute();
-int    dofirststage(u_char graphdebug, int debug_netnum);
-int    dosecondstage(u_char graphdebug, u_char singlestep);
-int    route_net_ripup(NET net, u_char graphdebug);
-
-void   read_lef(char *filename);
 void   read_def(char *filename);
 int    write_def(char *filename);
 
-int    doroute(NET net, u_char stage, u_char graphdebug);
-int    route_setup(struct routeinfo_ *iroute, u_char stage);
-int    next_route_setup(struct routeinfo_ *iroute, u_char stage);
-int    route_segs(struct routeinfo_ *iroute, u_char stage, u_char graphdebug);
-
-ROUTE  createemptyroute();
-void   emit_routes(char *filename, double oscale, int iscale);
-int    set_routes_to_net(NET net, int newflags, POINT *pushlist, SEG bbox,
-		u_char stage);
-int    set_route_to_net(NET net, ROUTE rt, int newflags, POINT *pushlist,
-		SEG bbox, u_char stage);
-
-void   initMask();
-void   createMask(NET net, u_char slack, u_char halo);
-void   fillMask(u_char value);
-
-void   pathstart(FILE *cmd, int layer, int x, int y, u_char special, double oscale,
-			double invscale, u_char horizontal);
-void   pathto(FILE *cmd, int x, int y, int horizontal, int lastx, int lasty,
-			double invscale);
-void   pathvia(FILE *cmd, int layer, int x, int y, int lastx, int lasty,
-			int gridx, int gridy, double invscale);
-
-void   highlight_starts(POINT glist);
-void   highlight_source();
-void   highlight_dest();
-void   highlight(int, int);
-int    recalc_spacing();
-void   draw_layout();
-
-void   helpmessage();
-
+char  *print_node_name(NODE node);
 void   print_nets(char *filename);
 void   print_routes(char *filename);
 void   print_nlgates(char *filename);
-char  *print_node_name(NODE node);
+void   print_net(NET net);
+void   print_gate(GATE gate);
 
-GATE   DefFindInstance(char *name);
-void   DefHashInstance(GATE gateginfo);
+int    dofirststage(u_char graphdebug, int debug_netnum);
+int    dosecondstage(u_char graphdebug, u_char singlestep);
+int    dothirdstage(u_char graphdebug, int debug_netnum);
+
+int    doroute(NET net, u_char stage, u_char graphdebug);
+NET    getnettoroute(int order);
+int    route_net_ripup(NET net, u_char graphdebug);
+
+#ifdef TCL_QROUTER
+void   tcl_printf(FILE *, const char *, ...);
+void   tcl_stdflush(FILE *);
+void   tcl_vprintf(FILE *, const char *, va_list);
+#endif
 
 #define QROUTER_H
 #endif 
