@@ -21,6 +21,9 @@
 #include "maze.h"
 #include "qconfig.h"
 #include "lef.h"
+#include "graphics.h"
+#include "node.h"
+#include "tkSimple.h"
 
 /* Global variables */
 
@@ -30,75 +33,128 @@ Tcl_Interp *consoleinterp;
 
 int stepnet = -1;
 
-/* This hash table speeds up DEF file reading */
-
-Tcl_HashTable InstanceTable;
-
 /* Command structure */
 
 typedef struct {
    const char	*cmdstr;
-   void		(*func)();
+   int		(*func)(ClientData clientData, Tcl_Interp *interp,
+                        int objc, Tcl_Obj *CONST objv[]);
 } cmdstruct;
 
 /* Forward declarations of commands */
 
-extern int Tk_SimpleObjCmd();
-extern int redraw();
-extern int qrouter_map();
-extern int qrouter_start();
-extern int qrouter_stage1();
-extern int qrouter_stage2();
-extern int qrouter_writedef();
-extern int qrouter_readdef();
-extern int qrouter_readlef();
-extern int qrouter_readconfig();
-extern int qrouter_failing();
-extern int qrouter_cost();
-extern int qrouter_tag();
-extern int qrouter_remove();
-extern int qrouter_obs();
-extern int qrouter_layerinfo();
-extern int qrouter_priority();
-extern int qrouter_ignore();
-extern int qrouter_via();
-extern int qrouter_resolution();
-extern int qrouter_congested();
-extern int qrouter_layers();
-extern int qrouter_passes();
-extern int qrouter_vdd();
-extern int qrouter_gnd();
-extern int qrouter_verbose();
-extern int QuitCallback();
+static int qrouter_map(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_start(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_stage1(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_stage2(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_stage3(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_writedef(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_readdef(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_readlef(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_readconfig(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_failing(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_cost(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_tag(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_remove(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_obs(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_layerinfo(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_priority(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_ignore(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_via(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_resolution(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_congested(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_layers(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_passes(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_vdd(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_gnd(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_verbose(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_print(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_quit(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
 
 static cmdstruct qrouter_commands[] =
 {
-   {"tag", (void *)qrouter_tag},
-   {"start", (void *)qrouter_start},
-   {"stage1", (void *)qrouter_stage1},
-   {"stage2", (void *)qrouter_stage2},
-   {"write_def", (void *)qrouter_writedef},
-   {"read_def", (void *)qrouter_readdef},
-   {"read_lef", (void *)qrouter_readlef},
-   {"read_config", (void *)qrouter_readconfig},
-   {"layer_info", (void *)qrouter_layerinfo},
-   {"obstruction", (void *)qrouter_obs},
-   {"ignore", (void *)qrouter_ignore},
-   {"priority", (void *)qrouter_priority},
-   {"via", (void *)qrouter_via},
-   {"resolution", (void *)qrouter_resolution},
-   {"congested", (void *)qrouter_congested},
-   {"layers", (void *)qrouter_layers},
-   {"passes", (void *)qrouter_passes},
-   {"vdd", (void *)qrouter_vdd},
-   {"gnd", (void *)qrouter_gnd},
-   {"failing", (void *)qrouter_failing},
-   {"remove", (void *)qrouter_remove},
-   {"cost", (void *)qrouter_cost},
-   {"map", (void *)qrouter_map},
-   {"verbose", (void *)qrouter_verbose},
-   {"redraw", (void *)redraw},
-   {"quit", (void *)QuitCallback},
+   {"tag", qrouter_tag},
+   {"start", qrouter_start},
+   {"stage1", qrouter_stage1},
+   {"stage2", qrouter_stage2},
+   {"stage3", qrouter_stage3},
+   {"write_def", qrouter_writedef},
+   {"read_def", qrouter_readdef},
+   {"read_lef", qrouter_readlef},
+   {"read_config", qrouter_readconfig},
+   {"layer_info", qrouter_layerinfo},
+   {"obstruction", qrouter_obs},
+   {"ignore", qrouter_ignore},
+   {"priority", qrouter_priority},
+   {"via", qrouter_via},
+   {"resolution", qrouter_resolution},
+   {"congested", qrouter_congested},
+   {"layers", qrouter_layers},
+   {"passes", qrouter_passes},
+   {"vdd", qrouter_vdd},
+   {"gnd", qrouter_gnd},
+   {"failing", qrouter_failing},
+   {"remove", qrouter_remove},
+   {"cost", qrouter_cost},
+   {"map", qrouter_map},
+   {"verbose", qrouter_verbose},
+   {"redraw", redraw},
+   {"print", qrouter_print},
+   {"quit", qrouter_quit},
    {"", NULL}  /* sentinel */
 };
 
@@ -156,7 +212,7 @@ void tcl_vprintf(FILE *f, const char *fmt, va_list args_in)
    va_list args;
    static char outstr[128] = "puts -nonewline std";
    char *outptr, *bigstr = NULL, *finalstr = NULL;
-   int i, nchars, result, escapes = 0, limit;
+   int i, nchars, escapes = 0;
 
    /* If we are printing an error message, we want to bring attention	*/
    /* to it by mapping the console window and raising it, as necessary.	*/
@@ -167,8 +223,8 @@ void tcl_vprintf(FILE *f, const char *fmt, va_list args_in)
       Tk_Window tkwind;
       tkwind = Tk_MainWindow(consoleinterp);
       if ((tkwind != NULL) && (!Tk_IsMapped(tkwind)))
-	 result = Tcl_Eval(consoleinterp, "wm deiconify .\n");
-      result = Tcl_Eval(consoleinterp, "raise .\n");
+         Tcl_Eval(consoleinterp, "wm deiconify .\n");
+      Tcl_Eval(consoleinterp, "raise .\n");
    }
 
    strcpy (outstr + 19, (f == stderr) ? "err \"" : "out \"");
@@ -218,7 +274,7 @@ void tcl_vprintf(FILE *f, const char *fmt, va_list args_in)
     *(outptr + 24 + nchars + escapes) = '\"';
     *(outptr + 25 + nchars + escapes) = '\0';
 
-    result = Tcl_Eval(consoleinterp, outptr);
+    Tcl_Eval(consoleinterp, outptr);
 
     if (bigstr != NULL) Tcl_Free(bigstr);
     if (finalstr != NULL) Tcl_Free(finalstr);
@@ -267,7 +323,7 @@ int QrouterTagCallback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
     Tcl_HashEntry *entry;
     Tcl_SavedResult state;
     int reset = FALSE;
-    int i, llen, cmdnum;
+    int i, llen;
 
     entry = Tcl_FindHashEntry(&QrouterTagTable, croot);
     postcmd = (entry) ? (char *)Tcl_GetHashValue(entry) : NULL;
@@ -410,8 +466,9 @@ int QrouterTagCallback(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
 /* Add a command tag callback					*/
 /*--------------------------------------------------------------*/
 
-int qrouter_tag(ClientData clientData,
-        Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_tag(ClientData clientData,
+            Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
 {
     Tcl_HashEntry *entry;
     char *hstring;
@@ -442,40 +499,6 @@ int qrouter_tag(ClientData clientData,
     return TCL_OK;
 }
 
-/*--------------------------------------------------------------*/
-/* Cell macro lookup based on the hash table			*/
-/*--------------------------------------------------------------*/
-
-GATE
-DefFindInstance(char *name)
-{
-    GATE ginst;
-    Tcl_HashEntry *entry;
-
-    entry = Tcl_FindHashEntry(&InstanceTable, name);
-    ginst = (entry) ? (GATE)Tcl_GetHashValue(entry) : NULL;
-    return ginst;
-}
-
-/*--------------------------------------------------------------*/
-/* Cell macro hash table generation				*/
-/* Given an instance record, create an entry in the hash table	*/
-/* for the instance name, with the record entry pointing to the	*/
-/* instance record.						*/
-/*--------------------------------------------------------------*/
-
-void
-DefHashInstance(GATE gateginfo)
-{
-    int new;
-    Tcl_HashEntry *entry;
-
-    entry = Tcl_CreateHashEntry(&InstanceTable,
-		gateginfo->gatename, &new);
-    if (entry != NULL)
-	Tcl_SetHashValue(entry, (ClientData)gateginfo);
-}
-
 /*--------------------------------------------------------------*/
 /* Initialization procedure for Tcl/Tk				*/
 /*--------------------------------------------------------------*/
@@ -485,7 +508,6 @@ Qrouter_Init(Tcl_Interp *interp)
 {
    int cmdidx;
    Tk_Window tktop;
-   char *tmp_s;
    char command[256];
    char version_string[20];
 
@@ -531,10 +553,6 @@ Qrouter_Init(Tcl_Interp *interp)
    if ((consoleinterp = Tcl_GetMaster(interp)) == NULL)
       consoleinterp = interp;
 
-   /* Initialize the macro hash table */
-
-   Tcl_InitHashTable(&InstanceTable, TCL_STRING_KEYS);
-
    /* Initialize the command tag table */
 
    Tcl_InitHashTable(&QrouterTagTable, TCL_STRING_KEYS);
@@ -546,10 +564,11 @@ Qrouter_Init(Tcl_Interp *interp)
 /* Command "start"					*/
 /*------------------------------------------------------*/
 
-int qrouter_start(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_start(ClientData clientData, Tcl_Interp *interp,
+              int objc, Tcl_Obj *CONST objv[])
 {
-    int i, j, result, argc;
+    int i, result, argc;
     char *scriptfile = NULL;
     char **argv;
 
@@ -588,6 +607,30 @@ int qrouter_start(ClientData clientData, Tcl_Interp *interp,
     return QrouterTagCallback(interp, objc, objv);
 }
 
+/*------------------------------------------------------*/
+/* Command: qrouter_quit				*/
+/*							*/
+/* Call tkcon's exit routine, which will make sure	*/
+/* the history file is updated before final exit.	*/
+/*------------------------------------------------------*/
+
+int
+qrouter_quit(ClientData clientData, Tcl_Interp *interp,
+	int objc, Tcl_Obj *CONST objv[])
+{
+    if (objc != 1) {
+	Tcl_WrongNumArgs(interp, 1, objv, "(no arguments)");
+	return TCL_ERROR;
+    }
+
+    if (consoleinterp == interp)
+	Tcl_Exit(TCL_OK);
+    else
+	Tcl_Eval(interp, "catch {tkcon eval exit}\n");
+
+    return TCL_OK;       /* Not reached */
+}
+
 /*------------------------------------------------------*/
 /* Command "map"					*/
 /*							*/
@@ -599,19 +642,22 @@ int qrouter_start(ClientData clientData, Tcl_Interp *interp,
 /*	map none	    route background is plain	*/
 /*	map routes	    draw routes over map	*/
 /*	map noroutes	    don't draw routes over map	*/
+/*	map unrouted	    draw unrouted nets over map	*/
+/*	map nounrouted	    don't draw unrouted nets	*/
 /*------------------------------------------------------*/
 
-int qrouter_map(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_map(ClientData clientData, Tcl_Interp *interp,
+            int objc, Tcl_Obj *CONST objv[])
 {
-    int i, idx, result;
+    int idx, result;
 
     static char *subCmds[] = {
 	"obstructions", "congestion", "estimate", "none",
-	"routes", "noroutes", NULL
+	"routes", "noroutes", "unrouted", "nounrouted", NULL
     };
     enum SubIdx {
-	ObsIdx, CongIdx, EstIdx, NoneIdx, RouteIdx, NoRouteIdx
+	ObsIdx, CongIdx, EstIdx, NoneIdx, RouteIdx, NoRouteIdx, UnroutedIdx, NoUnroutedIdx
     };
    
     if (objc != 2) {
@@ -652,19 +698,20 @@ int qrouter_map(ClientData clientData, Tcl_Interp *interp,
 	    }
 	    break;
 	case RouteIdx:
-	    if ((mapType & DRAW_MASK) != DRAW_ROUTES) {
-		mapType &= ~DRAW_MASK;
-		mapType |= DRAW_ROUTES;
-		draw_layout();
-	    }
+            mapType |= DRAW_ROUTES;
+            draw_layout();
 	    break;
 	case NoRouteIdx:
-	    if ((mapType & DRAW_MASK) != DRAW_NONE) {
-		mapType &= ~DRAW_MASK;
-		mapType |= DRAW_NONE;
-		draw_layout();
-	    }
+            mapType &= ~DRAW_ROUTES;
+            draw_layout();
 	    break;
+        case UnroutedIdx:
+            mapType |= DRAW_UNROUTED;
+            draw_layout();
+            break;
+        case NoUnroutedIdx:
+            mapType &= ~DRAW_UNROUTED;
+            draw_layout();
     }
     return QrouterTagCallback(interp, objc, objv);
 }
@@ -689,6 +736,40 @@ NET LookupNet(char *netname)
     return NULL;
 }
 
+/*------------------------------------------------------*/
+/* Find the net with number "number" in the list of	*/
+/* nets and return a pointer to it.			*/
+/*------------------------------------------------------*/
+
+NET LookupNetNr(int number)
+{
+    NET net;
+    int i;
+
+    for (i = 0; i < Numnets; i++) {
+       net = Nlnets[i];
+       if (net->netnum == number)
+	  return net;
+    }
+    return NULL;
+}
+
+/*------------------------------------------------------*/
+/* Find the gate instance named gatename and return a	*/
+/* pointer to it.					*/
+/*------------------------------------------------------*/
+
+GATE LookupGate(char *gatename)
+{
+    GATE gate;
+
+    for (gate = Nlgates; gate != NULL; gate = gate->next) {
+       if (!strcmp(gate->gatename, gatename))
+	  return gate;
+    }
+    return NULL;
+}
+
 /*------------------------------------------------------*/
 /* Command "stage1"					*/
 /*							*/
@@ -717,12 +798,13 @@ NET LookupNet(char *netname)
 /*  stage1 force	Force a terminal to be routable	*/
 /*------------------------------------------------------*/
 
-int qrouter_stage1(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_stage1(ClientData clientData, Tcl_Interp *interp,
+               int objc, Tcl_Obj *CONST objv[])
 {
     u_char dodebug;
     u_char dostep;
-    int i, idx, idx2, val, result, failcount;
+    int i, idx, idx2, val, result, failcount = 0;
     NET net = NULL;
 
     static char *subCmds[] = {
@@ -825,6 +907,24 @@ int qrouter_stage1(ClientData clientData, Tcl_Interp *interp,
 	if ((net != NULL) && (net->netnodes != NULL)) {
 	    result = doroute(net, (u_char)0, dodebug);
 	    failcount = (result == 0) ? 0 : 1;
+
+	    /* Remove from FailedNets list if routing	*/
+	    /* was successful				*/
+
+	    if (result == 0 && FailedNets != NULL) {
+		NETLIST fnet, lnet = NULL;
+		for (fnet = FailedNets; fnet != NULL; fnet = fnet->next) {
+		    if (fnet->net == net) {
+			if (lnet == NULL)
+			    FailedNets = fnet->next;
+			else
+			    lnet->next = fnet->next;
+			free(fnet);
+			break;
+		    }
+		    lnet = fnet;
+		}
+	    }
 	}
     }
     Tcl_SetObjResult(interp, Tcl_NewIntObj(failcount));
@@ -865,8 +965,9 @@ int qrouter_stage1(ClientData clientData, Tcl_Interp *interp,
 /*  stage2 tries <n>	Keep trying n additional times	*/
 /*------------------------------------------------------*/
 
-int qrouter_stage2(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_stage2(ClientData clientData, Tcl_Interp *interp,
+               int objc, Tcl_Obj *CONST objv[])
 {
     u_char dodebug;
     u_char dostep;
@@ -997,6 +1098,169 @@ int qrouter_stage2(ClientData clientData, Tcl_Interp *interp,
     return QrouterTagCallback(interp, objc, objv);
 }
 
+/*------------------------------------------------------*/
+/* Command "stage3"					*/
+/*							*/
+/* Execute stage3 routing.  This works through the	*/
+/* entire netlist, ripping up each route in turn and	*/
+/* re-routing it.  					*/
+/*							*/
+/* The interpreter result is set to the number of	*/
+/* failed routes at the end of the first stage.		*/
+/*							*/
+/* Options:						*/
+/*							*/
+/*  stage3 debug	Draw the area being searched in	*/
+/*			real-time.  This slows down the	*/
+/*			algorithm and is intended only	*/
+/*			for diagnostic use.		*/
+/*  stage3 step		Single-step stage three.	*/
+/*  stage3 mask none	Don't limit the search area	*/
+/*  stage3 mask auto	Select the mask automatically	*/
+/*  stage3 mask bbox	Use the net bbox as a mask	*/
+/*  stage3 mask <value> Set the mask size to <value>,	*/
+/*			an integer typ. 0 and up.	*/
+/*  stage3 route <net>	Route net named <net> only.	*/
+/*							*/
+/*  stage3 force	Force a terminal to be routable	*/
+/*------------------------------------------------------*/
+
+static int
+qrouter_stage3(ClientData clientData, Tcl_Interp *interp,
+               int objc, Tcl_Obj *CONST objv[])
+{
+    u_char dodebug;
+    u_char dostep;
+    int i, idx, idx2, val, result, failcount = 0;
+    NET net = NULL;
+
+    static char *subCmds[] = {
+	"debug", "mask", "route", "force", "step", NULL
+    };
+    enum SubIdx {
+	DebugIdx, MaskIdx, RouteIdx, ForceIdx, StepIdx
+    };
+   
+    static char *maskSubCmds[] = {
+	"none", "auto", "bbox", NULL
+    };
+    enum maskSubIdx {
+	NoneIdx, AutoIdx, BboxIdx
+    };
+
+    // Command defaults
+
+    dodebug = FALSE;
+    dostep = FALSE;
+    maskMode = MASK_AUTO;	// Mask mode is auto unless specified
+    forceRoutable = FALSE;	// Don't force unless specified
+
+    if (objc >= 2) {
+	for (i = 1; i < objc; i++) {
+
+	    if ((result = Tcl_GetIndexFromObj(interp, objv[i],
+			(CONST84 char **)subCmds, "option", 0, &idx))
+			!= TCL_OK)
+		return result;
+
+	    switch (idx) {
+		case DebugIdx:
+		    dodebug = TRUE;
+		    break;
+
+		case StepIdx:
+		    dostep = TRUE;
+		    break;
+
+		case ForceIdx:
+		    forceRoutable = TRUE;
+		    break;
+	
+		case RouteIdx:
+		    if (i >= objc - 1) {
+			Tcl_WrongNumArgs(interp, 0, objv, "route ?net?");
+			return TCL_ERROR;
+		    }
+		    i++;
+		    net = LookupNet(Tcl_GetString(objv[i]));
+		    if (net == NULL) {
+			Tcl_SetResult(interp, "No such net", NULL);
+			return TCL_ERROR;
+		    }
+		    break;
+
+		case MaskIdx:
+		    if (i >= objc - 1) {
+			Tcl_WrongNumArgs(interp, 0, objv, "mask ?type?");
+			return TCL_ERROR;
+		    }
+		    i++;
+		    if ((result = Tcl_GetIndexFromObj(interp, objv[i],
+				(CONST84 char **)maskSubCmds, "type", 0,
+				&idx2)) != TCL_OK) {
+			Tcl_ResetResult(interp);
+			result = Tcl_GetIntFromObj(interp, objv[i], &val);
+			if (result != TCL_OK) return result;
+			else if (val < 0 || val > 200) {
+			    Tcl_SetResult(interp, "Bad mask value", NULL);
+			    return TCL_ERROR;
+			}
+			maskMode = (u_char)val;
+		    }
+		    else {
+			switch(idx2) {
+			    case NoneIdx:
+				maskMode = MASK_NONE;
+				break;
+			    case AutoIdx:
+				maskMode = MASK_AUTO;
+				break;
+			    case BboxIdx:
+				maskMode = MASK_BBOX;
+				break;
+			}
+		    }
+		    break;
+	    }
+	}
+    }
+
+    if (dostep == FALSE) stepnet = -1;
+    else stepnet++;
+
+    if (net == NULL)
+	failcount = dothirdstage(dodebug, stepnet);
+    else {
+	if ((net != NULL) && (net->netnodes != NULL)) {
+	    result = doroute(net, (u_char)0, dodebug);
+	    failcount = (result == 0) ? 0 : 1;
+
+	    /* Remove from FailedNets list if routing	*/
+	    /* was successful				*/
+
+	    if (result == 0 && FailedNets != NULL) {
+		NETLIST fnet, lnet = NULL;
+		for (fnet = FailedNets; fnet != NULL; fnet = fnet->next) {
+		    if (fnet->net == net) {
+			if (lnet == NULL)
+			    FailedNets = fnet->next;
+			else
+			    lnet->next = fnet->next;
+			free(fnet);
+			break;
+		    }
+		    lnet = fnet;
+		}
+	    }
+	}
+    }
+    Tcl_SetObjResult(interp, Tcl_NewIntObj(failcount));
+
+    if (stepnet >= (Numnets - 1)) stepnet = -1;
+
+    return QrouterTagCallback(interp, objc, objv);
+}
+
 /*------------------------------------------------------*/
 /* Command "remove"					*/
 /*							*/
@@ -1011,8 +1275,9 @@ int qrouter_stage2(ClientData clientData, Tcl_Interp *interp,
 /*			the design.			*/
 /*------------------------------------------------------*/
 
-int qrouter_remove(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_remove(ClientData clientData, Tcl_Interp *interp,
+               int objc, Tcl_Obj *CONST objv[])
 {
     int result, idx, i;
     NET net;
@@ -1068,8 +1333,9 @@ int qrouter_remove(ClientData clientData, Tcl_Interp *interp,
 /*			as originally ordered		*/
 /*------------------------------------------------------*/
 
-int qrouter_failing(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_failing(ClientData clientData, Tcl_Interp *interp,
+                int objc, Tcl_Obj *CONST objv[])
 {
     Tcl_Obj *lobj;
     NETLIST nl, nlast;
@@ -1149,10 +1415,13 @@ int qrouter_failing(ClientData clientData, Tcl_Interp *interp,
 /* Command "read_lef"					*/
 /*------------------------------------------------------*/
 
-int qrouter_readlef(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_readlef(ClientData clientData, Tcl_Interp *interp,
+                int objc, Tcl_Obj *CONST objv[])
 {
     char *LEFfile;
+    int mscale;
+    int i;
 
     if (objc != 2) {
 	Tcl_SetResult(interp, "No LEF filename specified!", NULL);
@@ -1160,7 +1429,17 @@ int qrouter_readlef(ClientData clientData, Tcl_Interp *interp,
     }
     LEFfile = Tcl_GetString(objv[1]);
 
-    LefRead(LEFfile);
+    mscale = LefRead(LEFfile);
+    if (Scales.mscale < mscale) Scales.mscale = mscale;
+ 
+    for (i = 0; i < Num_layers; i++) {
+
+       /* Set Vert and PitchX from route info */
+
+       Vert[i] = (1 - LefGetRouteOrientation(i));
+       PitchX[i] = LefGetRoutePitch(i);
+    }
+
     post_config();
 
     return QrouterTagCallback(interp, objc, objv);
@@ -1170,8 +1449,9 @@ int qrouter_readlef(ClientData clientData, Tcl_Interp *interp,
 /* Command "read_def"					*/
 /*------------------------------------------------------*/
 
-int qrouter_readdef(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_readdef(ClientData clientData, Tcl_Interp *interp,
+                int objc, Tcl_Obj *CONST objv[])
 {
     if ((DEFfilename[0] == '\0') && (objc != 2)) {
 	Tcl_SetResult(interp, "No DEF filename specified!", NULL);
@@ -1192,8 +1472,9 @@ int qrouter_readdef(ClientData clientData, Tcl_Interp *interp,
 /* Command "write_def"					*/
 /*------------------------------------------------------*/
 
-int qrouter_writedef(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_writedef(ClientData clientData, Tcl_Interp *interp,
+                 int objc, Tcl_Obj *CONST objv[])
 {
     char *DEFoutfile = NULL;
 
@@ -1212,8 +1493,9 @@ int qrouter_writedef(ClientData clientData, Tcl_Interp *interp,
 /* Command "read_config"				*/
 /*------------------------------------------------------*/
 
-int qrouter_readconfig(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_readconfig(ClientData clientData, Tcl_Interp *interp,
+                   int objc, Tcl_Obj *CONST objv[])
 {
     FILE *configFILE;
     char *configname = NULL;
@@ -1249,8 +1531,9 @@ int qrouter_readconfig(ClientData clientData, Tcl_Interp *interp,
 /*	obstruction <xmin> <ymin> <xmax> <ymax> <layer>	*/
 /*------------------------------------------------------*/
 
-int qrouter_obs(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_obs(ClientData clientData, Tcl_Interp *interp,
+            int objc, Tcl_Obj *CONST objv[])
 {
     Tcl_Obj *lobj;
     Tcl_Obj *oobj;
@@ -1322,8 +1605,9 @@ int qrouter_obs(ClientData clientData, Tcl_Interp *interp,
 /*	ignore [<net> ...]				*/
 /*------------------------------------------------------*/
 
-int qrouter_ignore(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_ignore(ClientData clientData, Tcl_Interp *interp,
+               int objc, Tcl_Obj *CONST objv[])
 {
     int i;
     NET net;
@@ -1365,8 +1649,9 @@ int qrouter_ignore(ClientData clientData, Tcl_Interp *interp,
 /*	priority [<net> ...]				*/
 /*------------------------------------------------------*/
 
-int qrouter_priority(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_priority(ClientData clientData, Tcl_Interp *interp,
+                 int objc, Tcl_Obj *CONST objv[])
 {
     int i;
     char *netname;
@@ -1405,7 +1690,7 @@ int qrouter_priority(ClientData clientData, Tcl_Interp *interp,
 		}
 	    }
 	}
-	create_netorder(0, NULL);
+	create_netorder(0);
     }
     return QrouterTagCallback(interp, objc, objv);
 }
@@ -1444,11 +1729,12 @@ int qrouter_priority(ClientData clientData, Tcl_Interp *interp,
 /* No option is the same as option "layer_info all"	*/
 /*------------------------------------------------------*/
 
-int qrouter_layerinfo(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_layerinfo(ClientData clientData, Tcl_Interp *interp,
+                  int objc, Tcl_Obj *CONST objv[])
 {
     Tcl_Obj *lobj, *oobj;
-    int i, idx, idx2, val, result, layer;
+    int i, idx, idx2, val, result, layer = -1;
     char *layername;
 
     static char *subCmds[] = {
@@ -1572,6 +1858,7 @@ int qrouter_layerinfo(ClientData clientData, Tcl_Interp *interp,
 	    Tcl_SetObjResult(interp, lobj);
 	    break;
     }
+    return TCL_OK;
 }
 
 /*------------------------------------------------------*/
@@ -1596,8 +1883,9 @@ int qrouter_layerinfo(ClientData clientData, Tcl_Interp *interp,
 /*	via pattern [normal|inverted]			*/
 /*------------------------------------------------------*/
 
-int qrouter_via(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_via(ClientData clientData, Tcl_Interp *interp,
+            int objc, Tcl_Obj *CONST objv[])
 {
     int idx, idx2, result, value;
 
@@ -1693,8 +1981,9 @@ int qrouter_via(ClientData clientData, Tcl_Interp *interp,
 /*	resolution [<value>]				*/
 /*------------------------------------------------------*/
 
-int qrouter_verbose(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_verbose(ClientData clientData, Tcl_Interp *interp,
+                int objc, Tcl_Obj *CONST objv[])
 {
     int result, value;
 
@@ -1735,8 +2024,9 @@ int qrouter_verbose(ClientData clientData, Tcl_Interp *interp,
 /*	resolution [<value>]				*/
 /*------------------------------------------------------*/
 
-int qrouter_resolution(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_resolution(ClientData clientData, Tcl_Interp *interp,
+                   int objc, Tcl_Obj *CONST objv[])
 {
     int result, value;
 
@@ -1775,8 +2065,9 @@ int qrouter_resolution(ClientData clientData, Tcl_Interp *interp,
 /*	layers [<number>]				*/
 /*------------------------------------------------------*/
 
-int qrouter_layers(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_layers(ClientData clientData, Tcl_Interp *interp,
+               int objc, Tcl_Obj *CONST objv[])
 {
     int result, value;
 
@@ -1816,8 +2107,9 @@ int qrouter_layers(ClientData clientData, Tcl_Interp *interp,
 /*	passes [<number>]				*/
 /*------------------------------------------------------*/
 
-int qrouter_passes(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_passes(ClientData clientData, Tcl_Interp *interp,
+               int objc, Tcl_Obj *CONST objv[])
 {
     int result, value;
 
@@ -1852,8 +2144,9 @@ int qrouter_passes(ClientData clientData, Tcl_Interp *interp,
 /*	vdd [<name>]					*/
 /*------------------------------------------------------*/
 
-int qrouter_vdd(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_vdd(ClientData clientData, Tcl_Interp *interp,
+            int objc, Tcl_Obj *CONST objv[])
 {
     if (objc == 1) {
 	if (vddnet == NULL)
@@ -1884,8 +2177,9 @@ int qrouter_vdd(ClientData clientData, Tcl_Interp *interp,
 /*	gnd [<name>]					*/
 /*------------------------------------------------------*/
 
-int qrouter_gnd(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_gnd(ClientData clientData, Tcl_Interp *interp,
+            int objc, Tcl_Obj *CONST objv[])
 {
     if (objc == 1) {
 	if (gndnet == NULL)
@@ -1916,20 +2210,22 @@ int qrouter_gnd(ClientData clientData, Tcl_Interp *interp,
 /*	cost jog					*/
 /*	cost crossover					*/
 /*	cost block					*/
+/*	cost offset					*/
 /*	cost conflict					*/
 /*------------------------------------------------------*/
 
-int qrouter_cost(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_cost(ClientData clientData, Tcl_Interp *interp,
+             int objc, Tcl_Obj *CONST objv[])
 {
     int idx, result, value;
 
     static char *subCmds[] = {
 	"segment", "via", "jog", "crossover",
-	"block", "conflict", NULL
+	"block", "offset", "conflict", NULL
     };
     enum SubIdx {
-	SegIdx, ViaIdx, JogIdx, XOverIdx, BlockIdx, ConflictIdx
+	SegIdx, ViaIdx, JogIdx, XOverIdx, BlockIdx, OffsetIdx, ConflictIdx
     };
    
     value = 0;
@@ -1937,7 +2233,9 @@ int qrouter_cost(ClientData clientData, Tcl_Interp *interp,
 	result = Tcl_GetIntFromObj(interp, objv[2], &value);
 	if (result != TCL_OK) return result;
 
-	if (value <= 0) {
+	// Disallow negative costs.
+
+	if (value < 0) {
 	    Tcl_SetResult(interp, "Bad cost value", NULL);
 	    return TCL_ERROR;
 	}
@@ -1951,6 +2249,20 @@ int qrouter_cost(ClientData clientData, Tcl_Interp *interp,
 		(CONST84 char **)subCmds, "option", 0, &idx)) != TCL_OK)
 	return result;
 
+    switch (idx) {
+	case SegIdx:
+	case ViaIdx:
+	case ConflictIdx:
+
+	// Segment, via, and conflict costs must not be zero or
+	// bad things happen.
+
+	    if (value <= 0) {
+		Tcl_SetResult(interp, "Bad cost value", NULL);
+		return TCL_ERROR;
+	    }
+    }
+
     switch (idx) {
 	case SegIdx:
 	    if (objc == 2)
@@ -1980,6 +2292,13 @@ int qrouter_cost(ClientData clientData, Tcl_Interp *interp,
 		XverCost = value;
 	    break;
 
+	case OffsetIdx:
+	    if (objc == 2)
+		Tcl_SetObjResult(interp, Tcl_NewIntObj(OffsetCost));
+	    else
+		OffsetCost = value;
+	    break;
+
 	case BlockIdx:
 	    if (objc == 2)
 		Tcl_SetObjResult(interp, Tcl_NewIntObj(BlockCost));
@@ -2038,12 +2357,13 @@ int compcong(CLIST *a, CLIST *b)
 /*			gates in the design.		*/
 /*------------------------------------------------------*/
 
-int qrouter_congested(ClientData clientData, Tcl_Interp *interp,
-	int objc, Tcl_Obj *CONST objv[])
+static int
+qrouter_congested(ClientData clientData, Tcl_Interp *interp,
+                  int objc, Tcl_Obj *CONST objv[])
 {
     NET net;
     int i, x, y, nwidth, nheight, area, length;
-    int value, entries, numgates, result;
+    int entries, numgates, result;
     float density, *Congestion;
     CLIST *cgates, csrch;
     GATE gsrch;
@@ -2055,7 +2375,7 @@ int qrouter_congested(ClientData clientData, Tcl_Interp *interp,
 	result = Tcl_GetIntFromObj(interp, objv[1], &entries);
 	if (result != TCL_OK) return result;
 
-	if (value <= 0) {
+	if (entries <= 0) {
 	    Tcl_SetResult(interp, "List size must be > 0", NULL);
 	    return TCL_ERROR;
 	}
@@ -2083,7 +2403,9 @@ int qrouter_congested(ClientData clientData, Tcl_Interp *interp,
 
 	for (x = net->xmin; x < net->xmax; x++)
 	    for (y = net->ymin; y < net->ymax; y++)
-		Congestion[OGRID(x, y, 0)] += density;
+		if (x >= 0 && x < NumChannelsX[0] &&
+			y >= 0 && y < NumChannelsY[0])
+		    CONGEST(x, y) += density;
     }
 
     // Use instance bounding boxes to estimate average congestion
@@ -2117,7 +2439,7 @@ int qrouter_congested(ClientData clientData, Tcl_Interp *interp,
 	cavg = 0.0;
 	for (x = bbox.x1; x <= bbox.x2; x++) {
 	    for (y = bbox.y1; y <= bbox.y2; y++) {
-		cavg += Congestion[OGRID(x, y, 0)];
+		cavg += CONGEST(x, y);
 	    }
 	}
 	cavg /= (bbox.x2 - bbox.x1 + 1);
@@ -2160,3 +2482,73 @@ int qrouter_congested(ClientData clientData, Tcl_Interp *interp,
 }
 
 /*------------------------------------------------------*/
+/* Command "print"					*/
+/*							*/
+/* print an object                          		*/
+/*							*/
+/* Options:						*/
+/*							*/
+/*	print net <netname> 				*/
+/*	print netnr <netnumber>				*/
+/*      print gate <gatename>				*/
+/*------------------------------------------------------*/
+
+static int
+qrouter_print(ClientData clientData, Tcl_Interp *interp,
+              int objc, Tcl_Obj *CONST objv[])
+{
+    int idx, result, value;
+    NET net;
+    GATE gate;
+
+    static char *subCmds[] = {
+	"net", "netnr", "gate", NULL
+    };
+    enum SubIdx {
+	NetIdx, NetNrIdx, GateIdx
+    };
+
+    value = 0;
+    if (objc != 3) {
+	Tcl_WrongNumArgs(interp, 1, objv, "option ?arg?");
+	return TCL_ERROR;
+    }
+
+    if ((result = Tcl_GetIndexFromObj(interp, objv[1],
+		(CONST84 char **)subCmds, "option", 0, &idx)) != TCL_OK)
+	return result;
+
+    switch (idx) {
+        case NetIdx:
+            net = LookupNet(Tcl_GetString(objv[2]));
+            if (net == NULL) {
+                Tcl_SetResult(interp, "Net not found", NULL);
+                return TCL_ERROR;
+            }
+            print_net(net);
+            break;
+
+        case NetNrIdx:
+            result = Tcl_GetIntFromObj(interp, objv[2], &value);
+            if (result != TCL_OK) return result;
+            net = LookupNetNr(value);
+            if (net == NULL) {
+                Tcl_SetResult(interp, "Net not found", NULL);
+                return TCL_ERROR;
+            }
+            print_net(net);
+            break;
+
+        case GateIdx:
+            gate = LookupGate(Tcl_GetString(objv[2]));
+            if (gate == NULL) {
+                Tcl_SetResult(interp, "Gate not found", NULL);
+                return TCL_ERROR;
+            }
+            print_gate(gate);
+    }
+
+    return QrouterTagCallback(interp, objc, objv);
+}
+
+/*------------------------------------------------------*/
